from pydantic import Field
from typing import Union, Annotated

from .allwise_p3as_mep import allwise_p3as_mep
from .neowiser_p1bs_psd import neowiser_p1bs_psd


TableType = Annotated[
    Union[allwise_p3as_mep, neowiser_p1bs_psd], Field(discriminator="name")
]



from pathlib import Path
import yaml

import numpy as np
from pydantic import BaseModel, model_validator

from .io import DownloadConfig
from .process import AmpelConfig, AmpelInterface


class TimewiseConfig(BaseModel):
    download: DownloadConfig
    ampel: AmpelConfig

    @classmethod
    def from_yaml(cls, path: str | Path):
        path = Path(path)
        assert path.exists(), f"{path} not found!"
        with path.open("r") as f:
            config_dict = yaml.safe_load(f)
        return cls.model_validate(config_dict)

    @model_validator(mode="after")
    def validate_query_original_id_key(self) -> "TimewiseConfig":
        unique_keys = np.unique([q.original_id_key for q in self.download.queries])
        assert len(unique_keys) == 1, (
            "Can not use different 'original_id_key' in queries!"
        )
        return self

    def build_ampel_interface(self) -> AmpelInterface:
        return self.ampel.build_interface(
            self.download.queries[0].original_id_key, self.download.input_csv
        )



import numpy as np
import pandas as pd
import numpy.typing as npt


def get_visit_map(
    mjd: npt.NDArray[np.float64] | pd.Series,
) -> npt.NDArray[np.int64]:
    """
    Create a map datapoint to visit

    :param mjd: the MJDs of the observations
    :type mjd: npt.NDArray[float]
    :returns: visit map
    :rtype: npt.ArrayLike
    """
    # -------------------------   find epoch intervals   -------------------------- #
    sorted_mjds = np.sort(mjd)
    epoch_bounds_mask = (sorted_mjds[1:] - sorted_mjds[:-1]) > 100
    epoch_bins = np.array(
        [mjd.min() * 0.99]  # this makes sure that the first datapoint gets selected
        + list(
            ((sorted_mjds[1:] + sorted_mjds[:-1]) / 2)[epoch_bounds_mask]
        )  # finding the middle between
        +
        # two visits
        [
            mjd.max() * 1.01
        ]  # this just makes sure that the last datapoint gets selected as well
    )

    visit_mask = np.digitize(mjd, epoch_bins) - 1
    return visit_mask



from typing import Literal, Dict, Any, Sequence, List, cast
from functools import partial
from pathlib import Path
import logging

import pandas as pd
import numpy as np
from numpy import typing as npt
from pydantic import BaseModel
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.markers import MarkerStyle
from matplotlib.transforms import Affine2D

from timewise.plot import plot_lightcurve, plot_panstarrs_cutout, plot_sdss_cutout
from timewise.plot.lightcurve import BAND_PLOT_COLORS
from timewise.process import keys
from timewise.util.visits import get_visit_map
from timewise.config import TimewiseConfig


logger = logging.getLogger(__name__)


class DiagnosticPlotter(BaseModel):
    cutout: Literal["sdss", "panstarrs"] = "panstarrs"
    band_colors: Dict[str, str] = BAND_PLOT_COLORS
    lum_key: str = keys.FLUX_EXT

    def plot_lightcurve(
        self,
        stacked_lightcurve: pd.DataFrame | None = None,
        raw_lightcurve: pd.DataFrame | None = None,
        ax: plt.Axes | None = None,
        **kwargs,
    ):
        return plot_lightcurve(
            lum_key=self.lum_key,
            stacked_lightcurve=stacked_lightcurve,
            raw_lightcurve=raw_lightcurve,
            ax=ax,
            colors=self.band_colors,
            **kwargs,
        )

    def plot_cutout(self, ra: float, dec: float, radius_arcsec: float, ax: plt.Axes):
        if self.cutout == "sdss":
            plot_cutout = plot_sdss_cutout
        elif self.cutout == "panstarrs":
            plot_cutout = partial(plot_panstarrs_cutout, plot_color_image=True)
        else:
            raise NotImplementedError  # should never happen
        return plot_cutout(ra=ra, dec=dec, arcsec=radius_arcsec, ax=ax)

    def make_plot(
        self,
        stacked_lightcurve: pd.DataFrame | None,
        raw_lightcurve: pd.DataFrame,
        labels: npt.ArrayLike,
        source_ra: float,
        source_dec: float,
        selected_indices: list[Any],
        highlight_radius: float | None = None,
    ) -> tuple[plt.Figure, Sequence[plt.Axes]]:
        fig, axs = plt.subplots(
            nrows=2, gridspec_kw={"height_ratios": [3, 2]}, figsize=(5, 8)
        )

        self.plot_cutout(ra=source_ra, dec=source_dec, ax=axs[0], radius_arcsec=20)

        selected_mask = raw_lightcurve.index.isin(selected_indices)
        plot_lightcurve(
            raw_lightcurve=raw_lightcurve[~selected_mask],
            lum_key=self.lum_key,
            ax=axs[-1],
            save=False,
            colors={"w1": "gray", "w2": "lightgray"},
            add_to_label=" ignored",
        )
        self.plot_lightcurve(
            stacked_lightcurve=stacked_lightcurve,
            raw_lightcurve=raw_lightcurve[selected_mask],
            ax=axs[-1],
            save=False,
        )

        # set markers for clusters
        markers_strings = list(Line2D.filled_markers) + [
            "$1$",
            "$2$",
            "$3$",
            "$4$",
            "$5$",
            "$6$",
            "$7$",
            "$8$",
            "$9$",
        ]
        markers_straight = [MarkerStyle(im) for im in markers_strings]
        rot = Affine2D().rotate_deg(180)
        markers_rotated = [MarkerStyle(im, transform=rot) for im in markers_strings]
        markers = markers_straight + markers_rotated

        # calculate ra and dec relative to center of cutout
        ra = (raw_lightcurve.ra - source_ra) * 3600
        dec = (raw_lightcurve.dec - source_dec) * 3600

        # get visit map
        visit_map = get_visit_map(raw_lightcurve.mjd)

        # for each visit plot the datapoints on the cutout
        # for each visit plot the datapoints on the cutout
        for visit in np.unique(visit_map):
            m = visit_map == visit
            label = str(visit)
            axs[0].plot(
                [],
                [],
                marker=markers[visit],
                label=label,
                mec="k",
                mew=1,
                mfc="none",
                ls="",
            )

            for im, mec, zorder in zip(
                [selected_mask, ~selected_mask], ["k", "none"], [1, 0]
            ):
                mask = m & im

                for i_label in np.unique(labels):
                    label_mask = labels == i_label
                    final_mask = mask & label_mask
                    datapoints_label = raw_lightcurve[final_mask]
                    color = f"C{i_label}" if i_label != -1 else "grey"

                    if ("sigra" in datapoints_label.columns) and (
                        "sigdec" in datapoints_label.columns
                    ):
                        has_sig = (
                            ~datapoints_label.sigra.isna()
                            & ~datapoints_label.sigdec.isna()
                        )
                        _ra = ra[final_mask]
                        _dec = dec[final_mask]

                        axs[0].errorbar(
                            _ra[has_sig],
                            _dec[has_sig],
                            xerr=datapoints_label.sigra[has_sig] / 3600,
                            yerr=datapoints_label.sigdec[has_sig] / 3600,
                            marker=markers[visit],
                            ls="",
                            color=color,
                            zorder=zorder,
                            ms=10,
                            mec=mec,
                            mew=0.1,
                        )
                        axs[0].scatter(
                            _ra[~has_sig],
                            _dec[~has_sig],
                            marker=markers[visit],
                            color=color,
                            zorder=zorder,
                            edgecolors=mec,
                            linewidths=0.1,
                        )
                    else:
                        axs[0].scatter(
                            ra[final_mask],
                            dec[final_mask],
                            marker=markers[visit],
                            color=color,
                            zorder=zorder,
                            edgecolors=mec,
                            linewidths=0.1,
                        )

        if highlight_radius:
            circle = plt.Circle(
                (0, 0),
                highlight_radius,
                color="g",
                fill=False,
                ls="-",
                lw=3,
                zorder=0,
            )
            axs[0].add_artist(circle)

        # formatting
        title = axs[0].get_title()
        axs[-1].set_ylabel("Apparent Vega Magnitude")
        axs[-1].grid(ls=":", alpha=0.5)
        axs[0].set_title("")
        axs[0].legend(
            ncol=5,
            bbox_to_anchor=(0, 1, 1, 0),
            loc="lower left",
            mode="expand",
            title=title,
        )
        axs[0].set_aspect(1, adjustable="box")

        return fig, axs


def make_plot(
    config_path: Path,
    cutout: Literal["sdss", "panstarrs"],
    indices: List[int],
    output_directory: Path,
):
    cfg = TimewiseConfig.from_yaml(config_path)
    ampel_interface = cfg.build_ampel_interface()
    input_data = pd.read_csv(cfg.download.input_csv).set_index(
        ampel_interface.orig_id_key
    )
    plotter = DiagnosticPlotter(cutout=cutout)
    for index in indices:
        stacked_lightcurve = ampel_interface.extract_stacked_lightcurve(stock_id=index)
        raw_lightcurve = ampel_interface.extract_datapoints(stock_id=index)
        selected_dp_ids = ampel_interface.extract_selected_datapoint_ids(stock_id=index)
        labels = [0] * len(raw_lightcurve)
        source = input_data.loc[index]
        ra: float = cast(float, source.ra)
        dec: float = cast(float, source.dec)

        fig, axs = plotter.make_plot(
            stacked_lightcurve=stacked_lightcurve,
            raw_lightcurve=raw_lightcurve,
            labels=labels,
            source_ra=ra,
            source_dec=dec,
            selected_indices=selected_dp_ids,
        )
        fn = output_directory / f"{index}.pdf"
        logger.info(f"Saving plot to {fn}")
        fig.savefig(fn)
        plt.close()



from pathlib import Path
import logging

from pydantic import BaseModel, model_validator

from .interface import AmpelInterface


logger = logging.getLogger(__name__)
DEFAULT_TEMPLATE_PATH = Path(__file__).parent / "template.yml"


class AmpelConfig(BaseModel):
    mongo_db_name: str
    template_path: Path = DEFAULT_TEMPLATE_PATH
    uri: str = "localhost:27017"
    # will default to <mongo_db_name>_input
    input_mongo_db_name: str = ""

    @model_validator(mode="after")  # type: ignore
    def default_input_db_name(self) -> "AmpelConfig":
        if not self.input_mongo_db_name:
            self.input_mongo_db_name = self.mongo_db_name + "_input"
        return self

    def build_interface(self, original_id_key: str, input_csv: Path) -> AmpelInterface:
        return AmpelInterface(
            mongo_db_name=self.mongo_db_name,
            orig_id_key=original_id_key,
            input_csv=input_csv,
            input_mongo_db_name=self.input_mongo_db_name,
            template_path=self.template_path,
            uri=self.uri,
        )



from pathlib import Path
from typing import List
import pandas as pd
from pydantic import BaseModel, Field, model_validator

from .download import Downloader
from ..query import QueryType
from ..backend import BackendType
from ..types import TYPE_MAP


class DownloadConfig(BaseModel):
    input_csv: Path
    chunk_size: int = 500_000
    max_concurrent_jobs: int = 4
    poll_interval: float = 10.0
    queries: List[QueryType] = Field(..., description="One or more queries per chunk")
    backend: BackendType = Field(..., discriminator="type")

    service_url: str = "https://irsa.ipac.caltech.edu/TAP"

    @model_validator(mode="after")
    def validate_input_csv_columns(self) -> "DownloadConfig":
        """Ensure that the input CSV contains all columns required by queries."""
        # only validate if the CSV actually exists
        if not self.input_csv.exists():
            raise ValueError(f"CSV file does not exist: {self.input_csv}")

        # read just the header and first 10 lines
        input_table = pd.read_csv(self.input_csv, nrows=10)

        missing_columns = set()
        wrong_dtype = set()
        for qc in self.queries:
            for col, dtype in qc.input_columns.items():
                if col not in input_table.columns:
                    missing_columns.add(col)
                else:
                    try:
                        input_table[col].astype(TYPE_MAP[dtype])
                    except Exception:
                        wrong_dtype.add(col)

        msg = f"CSV file {self.input_csv}: "
        if missing_columns:
            raise KeyError(msg + f"Missing required columns: {sorted(missing_columns)}")
        if wrong_dtype:
            raise TypeError(
                msg
                + f"Columns not convertable to right data type: {sorted(wrong_dtype)}"
            )

        return self

    def build_downloader(self) -> Downloader:
        return Downloader(
            service_url=self.service_url,
            input_csv=self.input_csv,
            chunk_size=self.chunk_size,
            backend=self.backend,
            queries=self.queries,
            max_concurrent_jobs=self.max_concurrent_jobs,
            poll_interval=self.poll_interval,
        )



import sys
from queue import Queue
from threading import Thread, Event
from typing import Any, Optional, Callable, Mapping


class ErrorQueue(Queue):
    """Queue subclass whose join() re-raises exceptions from worker threads."""

    def __init__(self, stop_event: Event, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_queue: Queue = Queue()
        self.stop_event = stop_event

    def report_error(self, exc_info):
        """Called by workers to push an exception into the error queue."""
        self.error_queue.put(exc_info)
        # Also decrement unfinished_tasks, so join() won't block forever
        with self.all_tasks_done:
            self.unfinished_tasks = max(0, self.unfinished_tasks - 1)
            self.all_tasks_done.notify_all()

    def join(self):
        """Wait until all tasks are done, or raise if a worker failed."""
        with self.all_tasks_done:
            while self.unfinished_tasks:
                if not self.error_queue.empty():
                    exc_info = self.error_queue.get()
                    self.stop_event.set()
                    raise exc_info[1].with_traceback(exc_info[2])
                self.all_tasks_done.wait()

    def raise_errors(self):
        """
        Raise the first worker exception, if any.
        """
        if not self.error_queue.empty():
            exc_info = self.error_queue.get()
            raise exc_info[1].with_traceback(exc_info[2])


class ExceptionSafeThread(Thread):
    """Thread subclass that reports uncaught exceptions to the ErrorQueue."""

    def __init__(
        self,
        error_queue: Any,
        *,
        group: Optional[Any] = None,
        target: Optional[Callable[..., Any]] = None,
        name: Optional[str] = None,
        args: tuple = (),
        kwargs: Optional[Mapping[str, Any]] = None,
        daemon: Optional[bool] = None,
    ):
        super().__init__(
            group=group,
            target=target,
            name=name,
            args=args,
            kwargs=kwargs,
            daemon=daemon,
        )
        self.error_queue = error_queue

    def run(self):
        try:
            super().run()
        except Exception:
            self.error_queue.report_error(sys.exc_info())



from typing import Dict
import warnings
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

from timewise.process import keys


BAND_PLOT_COLORS = {"w1": "r", "w2": "b"}


def plot_lightcurve(
    lum_key: str,
    stacked_lightcurve: pd.DataFrame | None = None,
    raw_lightcurve: pd.DataFrame | None = None,
    ax: plt.Axes | None = None,
    colors: Dict[str, str] | None = None,
    add_to_label: str = "",
    **kwargs,
) -> tuple[plt.Figure, plt.Axes]:
    assert (stacked_lightcurve is not None) or (raw_lightcurve is not None)

    if not colors:
        colors = BAND_PLOT_COLORS

    if not ax:
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = plt.gcf()

    for b in ["w1", "w2"]:
        try:
            if isinstance(stacked_lightcurve, pd.DataFrame):
                ul_mask_stacked = np.array(
                    stacked_lightcurve[f"{b}{lum_key}{keys.UPPER_LIMIT}"]
                ).astype(bool)
                ax.errorbar(
                    stacked_lightcurve[keys.MEAN + "_mjd"][~ul_mask_stacked],
                    stacked_lightcurve[f"{b}{keys.MEAN}{lum_key}"][~ul_mask_stacked],
                    yerr=stacked_lightcurve[f"{b}{lum_key}{keys.RMS}"][
                        ~ul_mask_stacked
                    ],
                    label=f"{b}{add_to_label} stacked",
                    ls="",
                    marker="s",
                    c=colors[b],
                    markersize=4,
                    markeredgecolor="k",
                    ecolor="k",
                    capsize=2,
                )
                ax.scatter(
                    stacked_lightcurve[keys.MEAN + "_mjd"][ul_mask_stacked],
                    stacked_lightcurve[f"{b}{keys.MEAN}{lum_key}"][ul_mask_stacked],
                    marker="v",
                    c=colors[b],
                    alpha=0.7,
                    s=2,
                )

            if isinstance(raw_lightcurve, pd.DataFrame):
                m = ~raw_lightcurve[f"{b}{lum_key}"].isna()
                ul_mask_raw = raw_lightcurve[f"{b}{keys.ERROR_EXT}{lum_key}"].isna()

                tot_m = m & ~ul_mask_raw
                if np.any(tot_m):
                    ax.errorbar(
                        raw_lightcurve.mjd[tot_m],
                        raw_lightcurve[f"{b}{lum_key}"][tot_m],
                        yerr=raw_lightcurve[f"{b}{keys.ERROR_EXT}{lum_key}"][tot_m],
                        label=f"{b}{add_to_label}",
                        ls="",
                        marker="o",
                        c=colors[b],
                        markersize=4,
                        alpha=0.3,
                    )

                single_ul_m = m & ul_mask_raw
                if np.any(single_ul_m):
                    label = (
                        f"{b}{add_to_label} upper limits" if not np.any(tot_m) else ""
                    )
                    ax.scatter(
                        raw_lightcurve.mjd[single_ul_m],
                        raw_lightcurve[f"{b}{lum_key}"][single_ul_m],
                        marker="d",
                        c=colors[b],
                        alpha=0.3,
                        s=1,
                        label=label,
                    )

        except KeyError as e:
            raise KeyError(f"Could not find brightness key {e}!")

    if lum_key == keys.MAG_EXT:
        ylim = ax.get_ylim()
        ax.set_ylim(max(ylim), min(ylim))

    ax.set_xlabel("MJD")
    ax.set_ylabel(lum_key)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message="No artists with labels found to put in legend",
            category=UserWarning,
        )
        ax.legend()

    return fig, ax



from typing import ClassVar, Dict, Type
from pydantic import BaseModel


class TableConfig(BaseModel):
    name: str
    columns_dtypes: ClassVar[Dict[str, Type]]
    ra_column: ClassVar[str]
    dec_column: ClassVar[str]



from typing import Iterator
from pathlib import Path
import numpy as np
from numpy import typing as npt
import pandas as pd
import logging

logger = logging.getLogger(__name__)


class Chunk:
    def __init__(
        self, chunk_id: int, indices: npt.ArrayLike, row_indices: npt.ArrayLike
    ):
        self.chunk_id = chunk_id
        self.indices = indices
        self.row_numbers = row_indices


class Chunker:
    def __init__(self, input_csv: Path, chunk_size: int):
        self.input_csv = input_csv
        self.chunk_size = chunk_size
        self._n_rows = self._count_rows()
        logger.debug(f"found {self._n_rows} rows in {self.input_csv}")

    def _count_rows(self) -> int:
        chunk = 1024 * 1024  # Process 1 MB at a time.
        f = np.memmap(self.input_csv)
        num_newlines = sum(
            np.sum(f[i : i + chunk] == ord("\n")) for i in range(0, len(f), chunk)
        )
        del f
        return num_newlines - 1  # one header row

    def __len__(self) -> int:
        return int(np.ceil(self._n_rows / self.chunk_size))

    def __iter__(self) -> Iterator[Chunk]:
        for chunk_id in range(len(self)):
            yield self.get_chunk(chunk_id)

    def get_chunk(self, chunk_id: int) -> Chunk:
        if chunk_id >= len(self):
            raise IndexError(f"Invalid chunk_id {chunk_id}")
        start = chunk_id * self.chunk_size
        stop = min(start + self.chunk_size, self._n_rows)
        indices = pd.read_csv(self.input_csv, skiprows=start, nrows=stop - start).index
        logger.debug(f"chunk {chunk_id}: from {start} to {stop}")
        return Chunk(chunk_id, indices, np.arange(start=start, stop=stop))



import time
import threading
import logging
from queue import Empty
from typing import Dict, Iterator, cast, Sequence
from itertools import product
from pathlib import Path
from datetime import datetime, timedelta

import pandas as pd
import numpy as np
from astropy.table import Table
from pyvo.utils.http import create_session

from .stable_tap import StableTAPService
from ..backend import BackendType
from ..types import TAPJobMeta, TaskID, TYPE_MAP
from ..query import QueryType
from ..query.base import Query
from ..util.error_threading import ErrorQueue, ExceptionSafeThread
from ..chunking import Chunker, Chunk


logger = logging.getLogger(__name__)


class Downloader:
    def __init__(
        self,
        service_url: str,
        input_csv: Path,
        chunk_size: int,
        backend: BackendType,
        queries: list[QueryType],
        max_concurrent_jobs: int,
        poll_interval: float,
    ):
        self.backend = backend
        self.queries = queries
        self.input_csv = input_csv
        self.max_concurrent_jobs = max_concurrent_jobs
        self.poll_interval = poll_interval

        # ----------------------------
        # concurrency setup
        # ----------------------------
        # Shared state
        self.job_lock = threading.Lock()
        # (chunk_id, query_hash) -> job meta
        self.jobs: Dict[TaskID, TAPJobMeta] = {}

        self.stop_event = threading.Event()
        self.submit_queue: ErrorQueue = ErrorQueue(stop_event=self.stop_event)
        self.submit_thread = ExceptionSafeThread(
            error_queue=self.submit_queue, target=self._submission_worker, daemon=True
        )
        self.poll_thread = ExceptionSafeThread(
            error_queue=self.submit_queue, target=self._polling_worker, daemon=True
        )
        self.all_chunks_queued = False
        self.all_chunks_submitted = False

        # ----------------------------
        # TAP setup
        # ----------------------------
        self.session = create_session()
        self.service: StableTAPService = StableTAPService(
            service_url, session=self.session
        )

        self.chunker = Chunker(input_csv=input_csv, chunk_size=chunk_size)

    # ----------------------------
    # helpers
    # ----------------------------
    @staticmethod
    def get_task_id(chunk: Chunk, query: Query) -> TaskID:
        return TaskID(
            namespace="download", key=f"chunk{chunk.chunk_id:04d}_{query.hash}"
        )

    def iter_tasks(self) -> Iterator[TaskID]:
        for chunk in self.chunker:
            for q in self.queries:
                yield self.get_task_id(chunk, q)

    def iter_tasks_per_chunk(self) -> Iterator[list[TaskID]]:
        for chunk in self.chunker:
            yield [self.get_task_id(chunk, q) for q in self.queries]

    def load_job_meta(self):
        backend = self.backend
        for task in self.iter_tasks():
            if backend.meta_exists(task):
                logger.debug(f"found job metadata {task}")
                if task not in self.jobs:
                    try:
                        jm = TAPJobMeta(**backend.load_meta(task))
                        logger.debug(f"loaded {jm}")
                        logger.debug(f"setting {task}")
                        with self.job_lock:
                            self.jobs[task] = jm
                    except Exception:
                        continue

    # ----------------------------
    # TAP submission and download
    # ----------------------------
    def get_chunk_data(self, chunk: Chunk) -> pd.DataFrame:
        start = (
            min(cast(Sequence[int], chunk.row_numbers)) + 1
        )  # plus one to always skip header line
        nrows = (
            max(cast(Sequence[int], chunk.row_numbers)) - start + 2
        )  # plus one: skip header, plus one:

        columns = list(pd.read_csv(self.input_csv, nrows=0).columns)
        return pd.read_csv(
            filepath_or_buffer=self.input_csv,
            skiprows=start,
            nrows=nrows,
            names=columns,
        )

    def submit_tap_job(self, query: Query, chunk: Chunk) -> TAPJobMeta:
        adql = query.adql
        chunk_df = self.get_chunk_data(chunk)

        assert all(chunk_df.index.isin(chunk.indices)), (
            "Some inputs loaded from wrong chunk!"
        )
        assert all(np.isin(chunk.indices, chunk_df.index)), (
            f"Some indices are missing in chunk {chunk.chunk_id}!"
        )
        logger.debug(f"loaded {len(chunk_df)} objects")

        try:
            upload = Table(
                {
                    key: np.array(chunk_df[key]).astype(TYPE_MAP[dtype])
                    for key, dtype in query.input_columns.items()
                }
            )
        except KeyError as e:
            print(chunk_df)
            raise KeyError(e)

        logger.debug(f"uploading {len(upload)} objects.")
        job = self.service.submit_job(adql, uploads={query.upload_name: upload})
        job.run()
        logger.debug(job.url)

        return TAPJobMeta(
            url=job.url,
            query=adql,
            query_config=query.model_dump(),
            input_length=len(chunk_df),
            submitted=str(datetime.now()),
            last_checked=str(datetime.now()),
            status=job.phase,
            completed_at="",
        )

    def check_job_status(self, job_meta: TAPJobMeta) -> str:
        status = self.service.get_job_from_url(url=job_meta["url"]).phase
        job_meta["last_checked"] = str(datetime.now())
        return status

    def download_job_result(self, job_meta: TAPJobMeta) -> Table:
        logger.info(f"downloading {job_meta['url']}")
        job = self.service.get_job_from_url(url=job_meta["url"])
        job.wait()
        return job.fetch_result().to_table()

    # ----------------------------
    # Submission thread
    # ----------------------------
    def _submission_worker(self):
        while not self.stop_event.is_set():
            try:
                chunk, query = self.submit_queue.get(timeout=1.0)  # type: Chunk, Query
            except Empty:
                if self.all_chunks_queued:
                    self.all_chunks_submitted = True
                    break
                continue

            # Wait until we have capacity
            while not self.stop_event.is_set():
                with self.job_lock:
                    running = sum(
                        1
                        for j in self.jobs.values()
                        if j.get("status") in ("QUEUED", "EXECUTING", "RUNNING")
                    )
                if running < self.max_concurrent_jobs:
                    break
                time.sleep(1.0)

            task = self.get_task_id(chunk, query)
            logger.info(f"submitting {task}")
            job_meta = self.submit_tap_job(query, chunk)
            self.backend.save_meta(task, job_meta)
            with self.job_lock:
                self.jobs[task] = job_meta

            self.submit_queue.task_done()

    # ----------------------------
    # Polling thread
    # ----------------------------
    def _polling_worker(self):
        logger.debug("starting polling worker")
        backend = self.backend
        while not self.stop_event.is_set():
            # reload job infos
            self.load_job_meta()

            with self.job_lock:
                items = list(self.jobs.items())

            for task, meta in items:  # type: TaskID, TAPJobMeta
                if meta.get("status") in ("COMPLETED", "ERROR", "ABORTED"):
                    logger.debug(f"{task} was already {meta['status']}")
                    continue

                status = self.check_job_status(meta)
                if status == "COMPLETED":
                    logger.info(f"completed {task}")
                    payload_table = self.download_job_result(meta)
                    logger.debug(payload_table.columns)
                    backend.save_data(task, payload_table)
                    meta["status"] = "COMPLETED"
                    meta["completed_at"] = str(datetime.now())
                    backend.save_meta(task, meta)
                    backend.mark_done(task)
                    with self.job_lock:
                        self.jobs[task] = meta
                elif status in ("ERROR", "ABORTED"):
                    logger.warning(f"failed {task}: {status}")
                    meta["status"] = status
                    with self.job_lock:
                        self.jobs[task] = meta
                    backend.save_meta(task, meta)
                else:
                    with self.job_lock:
                        self.jobs[task]["status"] = status
                        snapshot = self.jobs[task]
                    backend.save_meta(task, snapshot)

            if self.all_chunks_submitted:
                with self.job_lock:
                    all_done = (
                        all(
                            j.get("status") in ("COMPLETED", "ERROR", "ABORTED")
                            for j in self.jobs.values()
                        )
                        if len(self.jobs) > 0
                        else False
                    )
                if all_done:
                    logger.info("All tasks done! Exiting polling thread")
                    break

            logger.info(
                f"Next poll at {datetime.now() + timedelta(seconds=self.poll_interval)}s"
            )
            time.sleep(self.poll_interval)

    # ----------------------------
    # Main run loop
    # ----------------------------
    def run(self):
        # load existing job metadata
        self.load_job_meta()

        # start threads
        self.submit_thread.start()
        self.poll_thread.start()

        # enqueue all chunks & queries
        backend = self.backend
        for chunk, q in product(self.chunker, self.queries):
            task = self.get_task_id(chunk, q)

            # skip if the download is done, or the job is queued
            if backend.is_done(task) or (task in self.jobs):
                continue

            self.submit_queue.put((chunk, q))
        self.all_chunks_queued = True
        # wait until all jobs are submitted
        self.submit_queue.join()
        # wait for the submit thread
        self.submit_thread.join()
        # the polling thread will exit ones all results are downloaded
        self.poll_thread.join()
        # the stop event will stop also the submit thread
        self.stop_event.set()
        # if any thread exited with an error report it
        self.submit_queue.raise_errors()
        logger.info("Done running downloader!")



import json
import logging
from pathlib import Path
from typing import Any, Literal
from astropy.table import Table

from .base import Backend
from ..types import TaskID


logger = logging.getLogger(__name__)


class FileSystemBackend(Backend):
    type: Literal["filesystem"] = "filesystem"
    base_path: Path

    # ----------------------------
    # Helpers for paths
    # ----------------------------
    def _meta_path(self, task: TaskID) -> Path:
        return self.base_path / f"{task}.meta.json"

    def _marker_path(self, task: TaskID) -> Path:
        return self.base_path / f"{task}.ok"

    def _data_path(self, task: TaskID) -> Path:
        return self.base_path / f"{task}.fits"

    # ----------------------------
    # Metadata
    # ----------------------------
    def save_meta(self, task: TaskID, meta: dict[str, Any]) -> None:
        path = self._meta_path(task)
        tmp = path.with_suffix(".tmp")
        tmp.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"writing {path}")
        tmp.write_text(json.dumps(meta, indent=2))
        tmp.replace(path)

    def load_meta(self, task: TaskID) -> dict[str, Any] | None:
        path = self._meta_path(task)
        if not path.exists():
            return None
        return json.loads(path.read_text())

    def meta_exists(self, task: TaskID) -> bool:
        return self._meta_path(task).exists()

    # ----------------------------
    # Markers
    # ----------------------------
    def mark_done(self, task: TaskID) -> None:
        mp = self._marker_path(task)
        mp.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"writing {mp}")
        mp.write_text("done")

    def is_done(self, task: TaskID) -> bool:
        return self._marker_path(task).exists()

    # ----------------------------
    # Data
    # ----------------------------
    def save_data(self, task: TaskID, content: Table) -> None:
        path = self._data_path(task)
        tmp = path.with_suffix(".tmp")
        tmp.parent.mkdir(parents=True, exist_ok=True)
        logger.debug(f"writing {path}")
        content.write(tmp, format="fits")
        tmp.replace(path)

    def load_data(self, task: TaskID) -> Table:
        path = self._data_path(task)
        if not path.exists():
            raise FileNotFoundError(path)
        return Table.read(path, format="fits")

    def data_exists(self, task: TaskID) -> bool:
        return self._data_path(task).exists()



import logging
from typing import Literal, Dict

from .base import Query

logger = logging.getLogger(__name__)


class PositionalQuery(Query):
    type: Literal["positional"] = "positional"
    radius_arcsec: float

    @property
    def input_columns(self) -> Dict[str, str]:
        return {"ra": "float", "dec": "float", self.original_id_key: "int"}

    def build(self) -> str:
        logger.debug(f"constructing positional query for {self.table.name}")

        q = "SELECT \n\t"
        for k in self.columns:
            q += f"{self.table.name}.{k}, "
        q += f"\n\tmine.{self.original_id_key} \n"
        q += f"FROM\n\tTAP_UPLOAD.{self.upload_name} AS mine \n"
        q += f"RIGHT JOIN\n\t{self.table.name} \n"
        q += "WHERE \n"
        q += (
            f"\tCONTAINS(POINT('J2000',{self.table.name}.{self.table.ra_column},{self.table.name}.{self.table.dec_column}),"
            f"CIRCLE('J2000',mine.ra,mine.dec,{self.radius_arcsec / 3600:.18f}))=1 "
        )

        if len(self.constraints) > 0:
            q += " AND (\n"
            for c in self.constraints:
                q += f"\t{self.table.name}.{c} AND \n"
            q = q.strip(" AND \n")
            q += "\t)"

        logger.debug(f"\n{q}")
        return q



from typing import Literal, ClassVar, Type, Dict
from .base import TableConfig


class allwise_p3as_mep(TableConfig):
    name: Literal["allwise_p3as_mep"] = "allwise_p3as_mep"
    columns_dtypes: ClassVar[Dict[str, Type]] = {
        "ra": float,
        "dec": float,
        "mjd": float,
        "cntr_mf": str,
        "w1mpro_ep": float,
        "w1sigmpro_ep": float,
        "w2mpro_ep": float,
        "w2sigmpro_ep": float,
        "w1flux_ep": float,
        "w1sigflux_ep": float,
        "w2flux_ep": float,
        "w2sigflux_ep": float,
    }
    ra_column: ClassVar[str] = "ra"
    dec_column: ClassVar[str] = "dec"



import logging
import hashlib
from pathlib import Path
from io import BytesIO
from PIL import Image

import requests
import matplotlib.pyplot as plt
import numpy
from astropy.table import Table
import astropy.config.paths


logger = logging.getLogger(__name__)


class PanSTARRSQueryError(Exception):
    pass


def load_cache_or_download(url):
    logger.debug(f"loading or downloading {url}")
    h = hashlib.md5(url.encode()).hexdigest()
    cache_dir = Path(astropy.config.paths.get_cache_dir())
    cache_file = cache_dir / (h + ".cache")
    logger.debug(f"cache file is {cache_file}")
    if not cache_file.is_file():
        logger.debug(f"downloading {url}")
        r = requests.get(url)
        with open(cache_file, "wb") as f:
            f.write(r.content)
        return r.content
    else:
        logger.debug(f"loading {cache_file}")
        with open(cache_file, "rb") as f:
            return f.read()


def annotate_not_available(ax):
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    x = sum(xlim) / 2
    y = sum(ylim) / 2
    logger.debug(f"annotate_not_available at {x}, {y}")
    ax.annotate(
        "Outside\nPanSTARRS\nFootprint",
        (x, y),
        color="red",
        ha="center",
        va="center",
        fontsize=10,
    )


def getimages(ra, dec, filters="grizy"):
    """Query ps1filenames.py service to get a list of images

    ra, dec = position in degrees
    size = image size in pixels (0.25 arcsec/pixel)
    filters = string with filters to include
    Returns a table with the results
    """

    service = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
    url = f"{service}?ra={ra}&dec={dec}&filters={filters}"
    content = load_cache_or_download(url)
    table = Table.read(content.decode(), format="ascii")
    return table


def geturl(
    ra, dec, size=240, output_size=None, filters="grizy", format="jpg", color=False
):
    """Get URL for images in the table

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filters = string with filters to include
    format = data format (options are "jpg", "png" or "fits")
    color = if True, creates a color image (only for jpg or png format).
            Default is return a list of URLs for single-filter grayscale images.
    Returns a string with the URL
    """

    if color and format == "fits":
        raise ValueError("color images are available only for jpg or png formats")
    if format not in ("jpg", "png", "fits"):
        raise ValueError("format must be one of jpg, png, fits")
    table = getimages(ra, dec, filters=filters)
    if len(table) == 0:
        raise PanSTARRSQueryError("No images available")
    url = (
        f"https://ps1images.stsci.edu/cgi-bin/fitscut.cgi?"
        f"ra={ra}&dec={dec}&size={size}&format={format}"
    )
    if output_size:
        url = url + "&output_size={}".format(output_size)
    # sort filters from red to blue
    flist = ["yzirg".find(x) for x in table["filter"]]
    table = table[numpy.argsort(flist)]
    if color:
        if len(table) > 3:
            # pick 3 filters
            table = table[[0, len(table) // 2, len(table) - 1]]
        for i, param in enumerate(["red", "green", "blue"]):
            url = url + "&{}={}".format(param, table["filename"][i])
    else:
        urlbase = url + "&red="
        url = []
        for filename in table["filename"]:
            url.append(urlbase + filename)
    return url


def getcolorim(ra, dec, size=240, output_size=None, filters="grizy", format="jpg"):
    """Get color image at a sky position

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filters = string with filters to include
    format = data format (options are "jpg", "png")
    Returns the image
    """

    if format not in ("jpg", "png"):
        raise ValueError("format must be jpg or png")
    url = geturl(
        ra,
        dec,
        size=size,
        filters=filters,
        output_size=output_size,
        format=format,
        color=True,
    )
    content = load_cache_or_download(url)
    im = Image.open(BytesIO(content))
    return im


def getgrayim(ra, dec, size=240, output_size=None, filter="g", format="jpg"):
    """Get grayscale image at a sky position

    ra, dec = position in degrees
    size = extracted image size in pixels (0.25 arcsec/pixel)
    output_size = output (display) image size in pixels (default = size).
                  output_size has no effect for fits format images.
    filter = string with filter to extract (one of grizy)
    format = data format (options are "jpg", "png")
    Returns the image
    """

    if format not in ("jpg", "png"):
        raise ValueError("format must be jpg or png")
    if filter not in list("grizy"):
        raise ValueError("filter must be one of grizy")
    url = geturl(
        ra, dec, size=size, filters=filter, output_size=output_size, format=format
    )
    content = load_cache_or_download(url[0])
    im = Image.open(BytesIO(content))
    return im


def plot_panstarrs_cutout(
    ra,
    dec,
    arcsec,
    interactive=False,
    fn=None,
    title=None,
    save=False,
    ax=False,
    plot_color_image=False,
    height=2.5,
):
    arcsec_per_px = 0.25
    ang_px = int(arcsec / arcsec_per_px)

    imshow_kwargs = {
        "origin": "upper",
        "extent": ([arcsec / 2, -arcsec / 2, -arcsec / 2, arcsec / 2]),
    }
    scatter_args = [0, 0]
    scatter_kwargs = {"marker": "x", "color": "red"}

    if not plot_color_image:
        filters = "grizy"
        if not ax:
            fig, axss = plt.subplots(
                2,
                len(filters),
                sharex="all",
                sharey="all",
                gridspec_kw={"wspace": 0, "hspace": 0, "height_ratios": [1, 8]},
                figsize=(height * 5, height),
            )
        else:
            fig = plt.gcf()
            axss = ax

        for j, fil in enumerate(list(filters)):
            axs = axss[1]
            try:
                im = getgrayim(ra, dec, size=ang_px, filter=fil)
                axs[j].imshow(im, cmap="gray", **imshow_kwargs)
            except PanSTARRSQueryError:
                axs[j].set_xlim(-arcsec / 2, arcsec / 2)
                axs[j].set_ylim(-arcsec / 2, arcsec / 2)
                annotate_not_available(axs[j])

            axs[j].scatter(*scatter_args, **scatter_kwargs)
            axs[j].set_title(fil)
            axss[0][j].axis("off")

    else:
        logger.debug("plotting color image")
        if not ax:
            fig, axss = plt.subplots(figsize=(height, height))
        else:
            fig = plt.gcf()
            axss = ax

        try:
            im = getcolorim(ra, dec, size=ang_px)
            axss.imshow(im, **imshow_kwargs)
        except PanSTARRSQueryError:
            axss.set_xlim(-arcsec / 2, arcsec / 2)
            axss.set_ylim(-arcsec / 2, arcsec / 2)
            annotate_not_available(axss)
        axss.scatter(*scatter_args, **scatter_kwargs)

    _this_title = title if title else f"{ra}_{dec}"
    si = "-" if dec > 0 else "+"
    ylabel = f"Dec {si} {abs(dec):.2f} [arcsec]"
    xlabel = f"RA - {ra:.2f} [arcsec]"
    try:
        axss.set_title(_this_title)
        axss.set_xlabel(xlabel)
        axss.set_ylabel(ylabel)
        axss.grid(ls=":", alpha=0.5)
    except AttributeError:  # in this case axss is an array
        fig.supylabel(ylabel)
        fig.supxlabel(xlabel)
        fig.suptitle(_this_title)
        for a in axss.flatten():
            a.grid(ls=":", alpha=0.5)

    if save:
        logger.info(f"saving under {fn}")
        fig.savefig(fn)

    if interactive:
        return fig, axss

    plt.close()



import abc
from typing import ClassVar, List
from pydantic import BaseModel
from hashlib import sha256

from ..tables import TableType


class Query(abc.ABC, BaseModel):
    type: str
    upload_name: ClassVar[str] = "mine"

    original_id_key: str = "orig_id"
    constraints: List[str] = [
        "nb < 2",
        "na < 1",
        "cc_flags like '00%'",
        "qi_fact >= 1",
        "saa_sep >= 5",
        "moon_masked like '00%'",
    ]
    columns: List[str]
    table: TableType

    @property
    @abc.abstractmethod
    def input_columns(self) -> dict[str, str]: ...

    @abc.abstractmethod
    def build(self) -> str: ...

    @property
    def adql(self) -> str:
        """ADQL string computed once per instance."""
        return self.build()

    @property
    def hash(self) -> str:
        return (
            self.type
            + "_"
            + self.table.name
            + "_"
            + sha256(self.adql.encode()).hexdigest()
        )



import logging
import backoff
import pyvo as vo
from xml.etree import ElementTree

from timewise.util.backoff import backoff_hndlr


logger = logging.getLogger(__name__)


class StableAsyncTAPJob(vo.dal.AsyncTAPJob):
    """
    Implements backoff for call of phase which otherwise breaks the code if there are connection issues.
    Also stores the response of TapQuery.submit() under self.submit_response for debugging
    """

    def __init__(self, url, *, session=None, delete=True):
        super(StableAsyncTAPJob, self).__init__(url, session=session, delete=delete)
        self.submit_response = None

    @classmethod
    def create(
        cls,
        baseurl,
        query,
        *,
        language="ADQL",
        maxrec=None,
        uploads=None,
        session=None,
        **keywords,
    ):
        """
        creates a async tap job on the server under ``baseurl``
        Raises requests.HTTPError if TAPQuery.submit() failes.

        Parameters
        ----------
        baseurl : str
            the TAP baseurl
        query : str
            the query string
        language : str
            specifies the query language, default ADQL.
            useful for services which allow to use the backend query language.
        maxrec : int
            the maximum records to return. defaults to the service default
        uploads : dict
            a mapping from table names to objects containing a votable
        session : object
           optional session to use for network requests
        """
        tapquery = vo.dal.TAPQuery(
            baseurl,
            query,
            mode="async",
            language=language,
            maxrec=maxrec,
            uploads=uploads,
            session=session,
            **keywords,
        )
        response = tapquery.submit()

        # check if the response is valid
        response.raise_for_status()

        # check if the response contains an error from the ADQL engine
        root = ElementTree.fromstring(response.content)
        info = root.find(".//v:INFO", {"v": "http://www.ivoa.net/xml/VOTable/v1.3"})
        if info and (info.attrib.get("value") == "ERROR"):
            raise vo.dal.DALQueryError(info.text.strip())

        # create the job instance
        job = cls(response.url, session=session)
        job._client_set_maxrec = maxrec
        job.submit_response = response

        return job

    @property
    @backoff.on_exception(
        backoff.expo,
        (vo.dal.DALServiceError, AttributeError),
        max_tries=50,
        on_backoff=backoff_hndlr,
    )
    def phase(self):
        return super(StableAsyncTAPJob, self).phase


class StableTAPService(vo.dal.TAPService):
    """
    Implements the StableAsyncTAPJob for job submission
    """

    @backoff.on_exception(
        backoff.expo,
        (vo.dal.DALServiceError, AttributeError, AssertionError),
        max_tries=5,
        on_backoff=backoff_hndlr,
    )
    def submit_job(
        self, query, *, language="ADQL", maxrec=None, uploads=None, **keywords
    ):
        job = StableAsyncTAPJob.create(
            self.baseurl,
            query,
            language=language,
            maxrec=maxrec,
            uploads=uploads,
            session=self._session,
            **keywords,
        )
        logger.debug(job.url)
        assert job.phase
        return job

    def get_job_from_url(self, url):
        return StableAsyncTAPJob(url, session=self._session)



import logging
from typing import Annotated, Literal, List
from pathlib import Path

import typer

from .config import TimewiseConfig
from .plot.diagnostic import make_plot

from rich.logging import RichHandler


app = typer.Typer(help="Timewsie CLI")

config_path_type = Annotated[
    Path, typer.Argument(help="Pipeline config file (YAML/JSON)")
]
ampel_config_path_type = Annotated[Path, typer.Argument(help="AMPEL config YAML")]


# --- Global callback (runs before every command) ---
@app.callback()
def main(
    ctx: typer.Context,
    log_level: str = typer.Option(
        "INFO",
        "--log-level",
        "-l",
        help="Logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
        case_sensitive=False,
    ),
):
    """Global options for all Timewise commands."""
    # Normalize log level
    level = getattr(logging, log_level.upper(), None)
    if not isinstance(level, int):
        raise typer.BadParameter(f"Invalid log level: {log_level}")

    # Rich logging
    logging.basicConfig(
        handlers=[RichHandler(rich_tracebacks=True, markup=True)],
    )
    logging.getLogger("timewise").setLevel(level)

    # Store log level in context for subcommands
    ctx.obj = {"log_level": level}


@app.command(help="Download WISE photometry from IRSA")
def download(
    config_path: config_path_type,
):
    TimewiseConfig.from_yaml(config_path).download.build_downloader().run()


@app.command(help="Prepares the AMPEL job file so AMPEL can be run manually")
def prepare_ampel(
    config_path: config_path_type,
):
    cfg = TimewiseConfig.from_yaml(config_path)
    ampel_interface = cfg.build_ampel_interface()
    p = ampel_interface.prepare(config_path)
    typer.echo(f"AMPEL job file: {p}")


@app.command(help="Processes the lightcurves using AMPEL")
def process(
    config_path: config_path_type,
    ampel_config_path: ampel_config_path_type,
):
    cfg = TimewiseConfig.from_yaml(config_path)
    ampel_interface = cfg.build_ampel_interface()
    ampel_interface.run(config_path, ampel_config_path)


@app.command(help="Write stacked lightcurves to disk")
def export(
    config_path: config_path_type,
    output_directory: Annotated[Path, typer.Argument(help="output directory")],
    indices: Annotated[
        list[int] | None,
        typer.Option(
            "-i", "--indices", help="Indices to export, defaults to all indices"
        ),
    ] = None,
):
    TimewiseConfig.from_yaml(config_path).build_ampel_interface().export_many(
        output_directory, indices
    )


@app.command(help="Run download, process and export")
def run_chain(
    config_path: config_path_type,
    ampel_config_path: ampel_config_path_type,
    output_directory: Annotated[Path, typer.Argument(help="output directory")],
    indices: Annotated[
        list[int] | None,
        typer.Option(
            "-i", "--indices", help="Indices to export, defaults to all indices"
        ),
    ] = None,
):
    download(config_path)
    process(config_path, ampel_config_path)
    export(config_path, output_directory, indices)


@app.command(help="Make diagnostic plots")
def plot(
    config_path: config_path_type,
    indices: Annotated[
        List[int],
        typer.Argument(help="Identifiers of the objects for which to create plots"),
    ],
    output_directory: Annotated[Path, typer.Argument(help="Output directory")],
    cutout: Annotated[
        Literal["sdss", "panstarrs"],
        typer.Option("-c", "--cutout", help="Which survey to use for cutouts"),
    ] = "panstarrs",
):
    make_plot(
        config_path, indices=indices, cutout=cutout, output_directory=output_directory
    )



import requests
import os
import getpass
import logging
import matplotlib.pyplot as plt
import backoff

from ..util.backoff import backoff_hndlr


logger = logging.getLogger(__name__)


def get_sdss_credentials():
    if not os.environ.get("SDSS_USERID"):
        os.environ["SDSS_USERID"] = input("Enter SDSS user ID:")
    if not os.environ.get("SDSS_USERPW"):
        os.environ["SDSS_USERPW"] = getpass.getpass("Enter SDSS password:")
    return os.environ["SDSS_USERID"], os.environ["SDSS_USERPW"]


def login_to_sciserver():
    try:
        from SciServer import Authentication
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            "Please install SciServer (https://github.com/sciserver/SciScript-Python) "
            "if you want to see SDSS cutouts!"
        )

    uid, pw = get_sdss_credentials()
    logger.debug(f"logging in to SciServer with username {uid}")
    Authentication.login(uid, pw)


@backoff.on_exception(
    backoff.expo, requests.RequestException, max_tries=50, on_backoff=backoff_hndlr
)
def get_cutout(*args, **kwargs):
    login_to_sciserver()
    from SciServer import SkyServer

    return SkyServer.getJpegImgCutout(*args, **kwargs)


def plot_sdss_cutout(
    ra,
    dec,
    arcsec=20,
    arcsec_per_px=0.1,
    interactive=False,
    fn=None,
    title=None,
    save=False,
    ax=False,
    height=2.5,
):
    ang_px = int(arcsec / arcsec_per_px)

    if not ax:
        fig, ax = plt.subplots(figsize=(height, height))
    else:
        fig = plt.gcf()

    try:
        im = get_cutout(ra, dec, scale=arcsec_per_px, height=ang_px, width=ang_px)
        ax.imshow(
            im,
            origin="upper",
            extent=(
                (
                    arcsec / 2,
                    -arcsec / 2,
                    -arcsec / 2,
                    arcsec / 2,
                )
            ),
            cmap="gray",
        )
        ax.scatter(0, 0, marker="x", color="red")

    except Exception as e:
        if "outside the SDSS footprint" in str(e):
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
            x = sum(xlim) / 2
            y = sum(ylim) / 2
            ax.annotate(
                "Outside SDSS Footprint",
                (x, y),
                color="red",
                ha="center",
                va="center",
                fontsize=20,
            )
        else:
            raise Exception(e)

    if title:
        ax.set_title(title)

    if save:
        logger.debug(f"saving under {fn}")
        fig.savefig(fn)

    if interactive:
        return fig, ax

    plt.close()



from typing import TypedDict, NamedTuple

from timewise.query import QueryType


class TAPJobMeta(TypedDict):
    url: str
    status: str
    submitted: str
    last_checked: str
    input_length: int
    query: str
    query_config: QueryType | dict
    completed_at: str


class TaskID(NamedTuple):
    """
    Generic identifier for a unit of work.
    Can be extended by Downloader/Processor as needed.
    """

    namespace: str  # e.g. "downloader", "processor"
    key: str  # unique string, e.g. "chunk_0001_q0" or "mask_2025-01-01"

    def __str__(self):
        return f"{self.namespace}_{self.key}"


TYPE_MAP = {"float": float, "str": str, "int": int}



import logging
from typing import cast, Dict, Any

from scipy import stats
import numpy as np
from numpy import typing as npt
import pandas as pd

from ..util.visits import get_visit_map
from timewise.process import keys


logger = logging.getLogger(__name__)


# zero points come from https://wise2.ipac.caltech.edu/docs/release/allsky/expsup/sec4_4h.html#conv2flux
# published in Jarret et al. (2011): https://ui.adsabs.harvard.edu/abs/2011ApJ...735..112J/abstract
MAGNITUDE_ZEROPOINTS: Dict[str, float] = {"w1": 20.752, "w2": 19.596}
# in Jy
FLUX_ZEROPOINTS = {"w1": 309.54, "w2": 171.787}


def calculate_epochs(
    f: pd.Series,
    e: pd.Series,
    visit_mask: npt.NDArray[np.int64],
    counts: npt.NDArray[np.int64],
    remove_outliers: bool,
    outlier_threshold: float,
    outlier_quantile: float,
    outlier_mask: npt.NDArray[np.bool_] | None = None,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.bool_],
    npt.NDArray[np.bool_],
    npt.NDArray[np.bool_],
    npt.NDArray[np.int64],
]:
    """
    Calculates the visits within a raw lightcurve.

    :param f: the fluxes
    :type f: np.array
    :param e: the flux errors
    :type e: np.array
    :param visit_mask: the visit mask
    :type visit_mask: np.array
    :param counts: the counts
    :type counts: np.array
    :param remove_outliers: whether to remove outliers
    :type remove_outliers: bool
    :param outlier_mask: the outlier mask
    :type outlier_mask: np.array
    :return: the epoch
    :rtype: float
    """

    if len(f) == 0:
        return (
            np.array([]),
            np.array([]),
            np.array([]),
            np.array([]),
            np.array([]),
            np.array([]),
        )

    u_lims = pd.isna(e)
    nan_mask = pd.isna(f)

    # ---------------------   remove outliers in the bins   ---------------------- #

    # if we do not want to clean outliers just set the threshold to infinity
    _outlier_threshold = np.inf if not remove_outliers else outlier_threshold

    # set up empty masks
    outlier_mask = cast(
        npt.NDArray[np.bool_],
        (np.array([False] * len(f)) if outlier_mask is None else outlier_mask),
    )
    median = np.full_like(counts, np.nan, dtype=float)
    u = np.full_like(counts, np.nan, dtype=float)
    use_mask = np.full_like(counts, False, dtype=bool)
    n_points = counts

    # set up dummy values for number of remaining outliers
    n_remaining_outlier = np.inf

    # ---------------------   flag upper limits   ---------------------- #
    bin_n_ulims: npt.NDArray[np.int64] = np.bincount(
        visit_mask, weights=u_lims, minlength=len(counts)
    )
    bin_ulim_bool = cast(npt.NDArray[np.bool_], (counts - bin_n_ulims) == 0)
    use_mask_ul = ~u_lims | (u_lims & bin_ulim_bool[visit_mask])

    n_loops = 0

    # recalculate uncertainty and median as long as no outliers left
    while n_remaining_outlier > 0:
        # make a mask of values to use
        use_mask = ~outlier_mask & use_mask_ul & ~nan_mask  # type: ignore[operator]
        n_points = np.bincount(visit_mask, weights=use_mask)
        zero_points_mask = cast(npt.NDArray[np.bool_], n_points == 0)

        # -------------------------   calculate median   ------------------------- #
        median = np.zeros_like(counts, dtype=float)
        visits_at_least_one_point = np.unique(visit_mask[~zero_points_mask[visit_mask]])
        visits_zero_points = np.unique(visit_mask[zero_points_mask[visit_mask]])
        median[visits_at_least_one_point] = np.array(
            [
                np.median(f[(visit_mask == i) & use_mask])
                for i in visits_at_least_one_point
            ]
        )
        median[visits_zero_points] = np.nan

        # median is NaN for visits with 0 detections, (i.e. detections in one band and not the other)
        # if median is NaN for other visits raise Error
        if np.any(np.isnan(median[n_points > 0])):
            nan_indices = np.where(np.isnan(median))[0]
            msg = ""
            for inan_index in nan_indices:
                nanf = f[visit_mask == inan_index]
                msg += f"median is nan for {inan_index}th bin\n{nanf}\n\n"
            raise ValueError(msg)

        # ---------------------   calculate uncertainty   ---------------------- #
        mean_deviation = np.bincount(
            visit_mask[use_mask],
            weights=(f[use_mask] - median[visit_mask[use_mask]]) ** 2,
            minlength=len(counts),
        )
        one_points_mask = n_points <= 1
        # calculate standard deviation
        std = np.zeros_like(counts, dtype=float)
        std[~one_points_mask] = (
            np.sqrt(mean_deviation[~one_points_mask])
            / (n_points[~one_points_mask] - 1)
            * stats.t.interval(0.68, df=n_points[~one_points_mask] - 1)[1]
            # for visits with small number of detections we have to correct according to the t distribution
        )
        std[one_points_mask] = -np.inf

        # calculate the propagated errors of the single exposure measurements
        single_exp_measurement_errors = np.sqrt(
            np.bincount(
                visit_mask[use_mask],
                weights=e[use_mask] ** 2,
                minlength=len(counts),
            )
        )
        e_meas = np.zeros_like(std, dtype=float)
        e_meas[~zero_points_mask] = (
            single_exp_measurement_errors[n_points > 0] / n_points[n_points > 0]
        )
        e_meas[zero_points_mask] = np.nan
        # take the maximum value of the measured single exposure errors and the standard deviation
        u = np.maximum(std, e_meas)

        # Estimate the spread of the flux.
        # To be robust against outliers, do that with quantiles instead of std
        qs = np.zeros_like(counts, dtype=float)
        qs[one_points_mask] = 1e-10
        visits_at_least_two_point = np.unique(visit_mask[~one_points_mask[visit_mask]])
        qs[visits_at_least_two_point] = np.array(
            [
                np.quantile(
                    abs(f[(visit_mask == i) & use_mask] - median[i]),
                    outlier_quantile,
                    method="interpolated_inverted_cdf",
                )
                for i in visits_at_least_two_point
            ]
        )

        # ---------------------   remove outliers in the bins   ---------------------- #
        remaining_outliers = (
            abs(median[visit_mask] - f) > _outlier_threshold * qs[visit_mask]
        ) & ~outlier_mask
        outlier_mask |= remaining_outliers
        n_remaining_outlier = sum(remaining_outliers) if remove_outliers else 0
        # setting remaining_outliers to 0 will exit the while loop

        n_loops += 1

        if n_loops > 20:
            raise Exception(f"{n_loops}!")

    return median, u, bin_ulim_bool, outlier_mask, use_mask, n_points


def stack_visits(
    lightcurve: pd.DataFrame,
    outlier_threshold: float,
    outlier_quantile: float,
    clean_outliers: bool = True,
):
    """
    Combine the data by visits of the satellite of one region in the sky.
    The visits typically consist of some tens of observations. The individual visits are separated by about
    six months.
    The mean flux for one visit is calculated by the weighted mean of the data.
    The error on that mean is calculated by the root-mean-squared and corrected by the t-value.
    Outliers per visit are identified if they are more than 100 times the rms away from the mean. These outliers
    are removed from the calculation of the mean and the error if self.clean_outliers_when_stacking is True.

    :param lightcurve: the raw lightcurve
    :type lightcurve: pandas.DataFrame
    :return: the stacked lightcurve
    :rtype: pandas.DataFrame
    """

    # -------------------------   create visit mask   -------------------------- #
    visit_map = get_visit_map(lightcurve.mjd)
    counts = np.bincount(visit_map)

    stacked_data: Dict[str, Any] = dict()

    # -------------------------   calculate mean mjd   -------------------------- #
    stacked_data["mean_mjd"] = np.bincount(visit_map, weights=lightcurve.mjd) / counts

    # -------------------------   loop through bands   -------------------------- #
    for b in ["w1", "w2"]:
        # loop through magnitude and flux and save the respective datapoints

        outlier_masks: Dict[str, Any] = dict()
        use_masks = dict()
        bin_ulim_bools = dict()

        for lum_ext in [keys.FLUX_EXT, keys.MAG_EXT]:
            f = lightcurve[f"{b}{lum_ext}"]
            e = lightcurve[f"{b}{keys.ERROR_EXT}{lum_ext}"]

            # we will flag outliers based on the flux only
            remove_outliers = lum_ext == keys.FLUX_EXT and clean_outliers
            outlier_mask = outlier_masks.get(keys.FLUX_EXT, None)

            mean, u, bin_ulim_bool, outlier_mask, use_mask, n_points = calculate_epochs(
                f,
                e,
                visit_map,
                counts,
                remove_outliers=remove_outliers,
                outlier_mask=outlier_mask,
                outlier_quantile=outlier_quantile,
                outlier_threshold=outlier_threshold,
            )
            n_outliers = np.sum(outlier_mask)

            if n_outliers > 0:
                logger.debug(
                    f"removed {n_outliers} outliers by brightness for {b} {lum_ext}"
                )

            stacked_data[f"{b}{keys.MEAN}{lum_ext}"] = mean
            stacked_data[f"{b}{lum_ext}{keys.RMS}"] = u
            stacked_data[f"{b}{lum_ext}{keys.UPPER_LIMIT}"] = bin_ulim_bool
            stacked_data[f"{b}{lum_ext}{keys.NPOINTS}"] = n_points

            outlier_masks[lum_ext] = outlier_mask
            use_masks[lum_ext] = use_mask
            bin_ulim_bools[lum_ext] = bin_ulim_bool

        # -------  calculate the zeropoints per exposure ------- #
        # this might look wrong since we use the flux mask on the magnitudes but it s right
        # for each flux measurement we need the corresponding magnitude to get the zeropoint
        mags = lightcurve[f"{b}{keys.MAG_EXT}"]
        inst_fluxes = lightcurve[f"{b}{keys.FLUX_EXT}"]
        pos_m = inst_fluxes > 0  # select only positive fluxes, i.e. detections
        zp_mask = pos_m & use_masks[keys.FLUX_EXT]

        # calculate zero points
        zps = np.zeros_like(inst_fluxes)
        zps[zp_mask] = mags[zp_mask] + 2.5 * np.log10(inst_fluxes[zp_mask])
        # find visits with no zeropoints
        n_valid_zps = np.bincount(visit_map, weights=zp_mask)
        at_least_one_valid_zp = n_valid_zps > 0
        # calculate the median zeropoint for each visit
        zps_median = np.zeros_like(n_valid_zps, dtype=float)
        zps_median[n_valid_zps > 0] = np.array(
            [
                np.median(zps[(visit_map == i) & zp_mask])
                for i in np.unique(visit_map[at_least_one_valid_zp[visit_map]])
            ]
        )
        # if there are only non-detections then fall back to default zeropoint
        zps_median[n_valid_zps == 0] = MAGNITUDE_ZEROPOINTS[b]
        # if the visit only has upper limits then use the fall-back zeropoint
        zps_median[bin_ulim_bools[keys.FLUX_EXT]] = MAGNITUDE_ZEROPOINTS[b]

        # ---------------   calculate flux density from instrument flux   ---------------- #
        # get the instrument flux [digital numbers], i.e. source count
        inst_fluxes_e = lightcurve[f"{b}{keys.ERROR_EXT}{keys.FLUX_EXT}"]

        # calculate the proportionality constant between flux density and source count
        mag_zp = FLUX_ZEROPOINTS[b] * 1e3  # in mJy
        flux_dens_const = mag_zp * 10 ** (-zps_median / 2.5)

        # calculate flux densities from instrument counts
        flux_densities = inst_fluxes * flux_dens_const[visit_map]
        flux_densities_e = inst_fluxes_e * flux_dens_const[visit_map]

        # bin flux densities
        mean_fd, u_fd, ul_fd, outlier_mask_fd, use_mask_fd, n_points_fd = (
            calculate_epochs(
                flux_densities,
                flux_densities_e,
                visit_map,
                counts,
                remove_outliers=False,
                outlier_mask=outlier_masks[keys.FLUX_EXT],
                outlier_threshold=outlier_threshold,
                outlier_quantile=outlier_quantile,
            )
        )
        stacked_data[f"{b}{keys.MEAN}{keys.FLUX_DENSITY_EXT}"] = mean_fd
        stacked_data[f"{b}{keys.FLUX_DENSITY_EXT}{keys.RMS}"] = u_fd
        stacked_data[f"{b}{keys.FLUX_DENSITY_EXT}{keys.UPPER_LIMIT}"] = ul_fd
        stacked_data[f"{b}{keys.FLUX_DENSITY_EXT}{keys.NPOINTS}"] = n_points_fd

    return pd.DataFrame(stacked_data)



from typing import Literal, ClassVar, Dict, Type
from .base import TableConfig


class neowiser_p1bs_psd(TableConfig):
    name: Literal["neowiser_p1bs_psd"] = "neowiser_p1bs_psd"
    columns_dtypes: ClassVar[Dict[str, Type]] = {
        "ra": float,
        "dec": float,
        "mjd": float,
        "allwise_cntr": str,
        "w1mpro": float,
        "w1sigmpro": float,
        "w2mpro": float,
        "w2sigmpro": float,
        "w1flux": float,
        "w1sigflux": float,
        "w2flux": float,
        "w2sigflux": float,
    }
    ra_column: ClassVar[str] = "ra"
    dec_column: ClassVar[str] = "dec"



import logging


logger = logging.getLogger(__name__)


def backoff_hndlr(details):
    logger.info(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )



import numpy as np
from pathlib import Path


def get_n_rows(path: str | Path):
    chunk = 1024 * 1024  # Process 1 MB at a time.
    f = np.memmap(path)
    num_newlines = sum(
        np.sum(f[i : i + chunk] == ord("\n")) for i in range(0, len(f), chunk)
    )
    del f
    return num_newlines



from pathlib import Path
import logging
from typing import Iterable, List, cast

import numpy as np
from numpy import typing as npt
import pandas as pd
from pymongo import MongoClient, ASCENDING
from pymongo.collection import Collection
from pymongo.database import Database
from ampel.cli.JobCommand import JobCommand
from ampel.types import DataPointId, StockId


logger = logging.getLogger(__name__)


class AmpelInterface:
    def __init__(
        self,
        mongo_db_name: str,
        orig_id_key: str,
        input_csv: Path,
        input_mongo_db_name: str,
        template_path: str | Path,
        uri: str,
    ):
        self.mongo_db_name = mongo_db_name
        self.orig_id_key = orig_id_key
        self.input_csv = input_csv
        self.input_mongo_db_name = input_mongo_db_name
        self.template_path = Path(template_path)
        self.uri = uri

    def import_input(self):
        # if collection already exists, assume import was already done
        if "input" in self.client[self.input_mongo_db_name].list_collection_names():
            logger.debug(
                f"'input' collection already exists in '{self.input_mongo_db_name}'."
            )
            return

        logger.debug(f"importing {self.input_csv} into {self.input_mongo_db_name}")
        col = self.client[self.input_mongo_db_name]["input"]

        # create an index from stock id
        col.create_index([(self.orig_id_key, ASCENDING)], unique=True)
        col.insert_many(pd.read_csv(self.input_csv).to_dict(orient="records"))

    def make_ampel_job_file(self, cfg_path: Path) -> Path:
        logger.debug(f"loading ampel job template from {self.template_path}")
        with self.template_path.open("r") as f:
            template = f.read()

        ampel_job = (
            template.replace("TIMEWISE_CONFIG_PATH", str(cfg_path))
            .replace("ORIGINAL_ID_KEY", self.orig_id_key)
            .replace("INPUT_MONGODB_NAME", self.input_mongo_db_name)
            .replace("MONGODB_NAME", self.mongo_db_name)
        )

        ampel_job_path = cfg_path.parent / f"{cfg_path.stem}_ampel_job.yml"
        logger.info(f"writing ampel job to {ampel_job_path}")
        with ampel_job_path.open("w") as f:
            f.write(ampel_job)

        return ampel_job_path

    def prepare(self, cfg_path: Path) -> Path:
        self.import_input()
        return self.make_ampel_job_file(cfg_path)

    def run(self, timewise_cfg_path: Path, ampel_config_path: Path):
        ampel_job_path = self.prepare(timewise_cfg_path)
        cmd = JobCommand()
        parser = cmd.get_parser()
        args = vars(
            parser.parse_args(
                ["--schema", str(ampel_job_path), "--config", str(ampel_config_path)]
            )
        )
        logger.debug(args)
        cmd.run(args, unknown_args=())

    @property
    def client(self) -> MongoClient:
        return MongoClient(self.uri)

    @property
    def db(self) -> Database:
        return self.client[self.mongo_db_name]

    @property
    def t0(self) -> Collection:
        return self.db["t0"]

    @property
    def t1(self) -> Collection:
        return self.db["t1"]

    @property
    def t2(self) -> Collection:
        return self.db["t2"]

    def extract_stacked_lightcurve(self, stock_id: StockId) -> pd.DataFrame:
        records = []
        for i, ic in enumerate(
            self.t2.find({"stock": stock_id, "unit": "T2StackVisits"})
        ):
            stock_id_str = str(stock_id)
            assert i == 0, f"More than one stacked lightcurve found for {stock_id_str}!"
            assert len(ic["body"]) == 1, (
                f"None or more than one stacking result found for {stock_id_str}!"
            )
            records = ic["body"][0]
        return pd.DataFrame(records)

    def extract_datapoints(self, stock_id: StockId) -> pd.DataFrame:
        records = []
        index = []
        for ic in self.t0.find({"stock": stock_id}):
            records.append(ic["body"])
            index.append(ic["id"])
        return pd.DataFrame(records, index=index)

    def extract_selected_datapoint_ids(self, stock_id: StockId) -> List[DataPointId]:
        d = self.t1.find_one({"stock": stock_id})
        if d is None:
            return []
        return d["dps"]

    def export_stacked_lightcurve(self, stock_id: StockId, filename: Path):
        logger.debug(f"Exporting stacked lightcurve for {str(stock_id)} to {filename}")
        self.extract_stacked_lightcurve(stock_id).to_csv(filename)

    def export_many(
        self, directory: Path, stock_ids: Iterable[StockId] | StockId | None = None
    ):
        if stock_ids is None:
            stock_ids = pd.read_csv(self.input_csv)[self.orig_id_key]
        directory.mkdir(exist_ok=True, parents=True)
        for s in np.atleast_1d(cast(npt.ArrayLike, stock_ids)):
            self.export_stacked_lightcurve(s.item(), directory / f"{s}.csv")



import abc
from typing import Any
from pydantic import BaseModel
from astropy.table import Table
from ..types import TaskID


class Backend(abc.ABC, BaseModel):
    type: str
    base_path: Any
    """
    Abstract persistence backend for jobs, results, and markers.
    Works with generic TaskIDs so it can be reused across Downloader/Processor.
    """

    # --- metadata ---
    @abc.abstractmethod
    def meta_exists(self, task: TaskID) -> bool: ...
    @abc.abstractmethod
    def save_meta(self, task: TaskID, meta: dict[str, Any]) -> None: ...
    @abc.abstractmethod
    def load_meta(self, task: TaskID) -> dict[str, Any] | None: ...

    # --- Markers ---
    @abc.abstractmethod
    def mark_done(self, task: TaskID) -> None: ...
    @abc.abstractmethod
    def is_done(self, task: TaskID) -> bool: ...

    # --- Data ---
    @abc.abstractmethod
    def save_data(self, task: TaskID, content: Table) -> None: ...
    @abc.abstractmethod
    def load_data(self, task: TaskID) -> Table: ...
    @abc.abstractmethod
    def data_exists(self, task: TaskID) -> bool: ...



from pydantic import Field
from typing import Union, Annotated, TypeAlias
from .positional import PositionalQuery

# Discriminated union of all query types
QueryType: TypeAlias = Annotated[Union[PositionalQuery], Field(discriminator="type")]
